#! /bin/sh
# This is a shell archive.  Remove anything before this line, then unpack
# it by saving it into a file and typing "sh file".  To overwrite existing
# files, type "sh file -c".  You can also feed this as standard input via
# unshar, or by typing "sh <file", e.g..  If this archive is complete, you
# will see the following message at the end:
#		"End of archive 2 (of 2)."
# Contents:  plscheme/doc/plscheme-1.tex plscheme/vm.scm
# Wrapped by oz@yunexus on Wed Mar 27 15:28:24 1991
PATH=/bin:/usr/bin:/usr/ucb ; export PATH
if test -f 'plscheme/doc/plscheme-1.tex' -a "${1}" != "-c" ; then 
  echo shar: Will not clobber existing file \"'plscheme/doc/plscheme-1.tex'\"
else
echo shar: Extracting \"'plscheme/doc/plscheme-1.tex'\" \(35366 characters\)
sed "s/^X//" >'plscheme/doc/plscheme-1.tex' <<'END_OF_FILE'
X%   p. 8 after the C[ exp, env, cont ], either omit the blank line in the
X% LaTeX source or put \noindent at the start of the following text.
X
X% Fix number of registers
X
X% Make environment notation consisten between compiler and
X% rewrite rules.
X
X% -*- Mode: TeX -*-
X% 6821 Handout 60.
X\documentstyle[11pt,6821,twoside]{article}
X
X\handout{70}{8 December 1988}
X\handouttitle{{\em New!  Improved!}\\Implementation of Scheme-like Languages, Part I}
X
X\input{macros}
X\input{commands}
X\newcommand{\rulename}[1]{\mbox{[{\it#1}\/]}}
X
X\begin{body}
X
X\note{This handout is an improved version of Handout~60 with numerous bug fixes.\footnote{Many 
Xthanks to Mark Sheldon and numerous students for pointing out the bugs.}}
X
X\section*{Introduction}
XIn this portion of 6.821 we shall discuss the implementation of Scheme-like
Xlanguages.  We shall use as an extended example a system called PLScheme, a
Xdialect of Scheme much like those we have previously considered in this
Xcourse.\footnote{The ``PL'' part of ``PLScheme'' is for ``Programming
XLanguages,'' the (non-numeric) name of this course.  PLScheme was written by
XJonathan Rees, whose aid in preparing this material the 6.821 staff gratefully
Xacknowledges.} PLScheme has all the usual functional features of the Scheme
Xdialects we have seen before; all side effects in PLScheme, however, 
Xare expressed using an ML-like mutable cell rather than the traditional 
X\code{set!}, \code{set-car!} and \code{set-cdr!}.
XIn this handout we describe aspects of the PLScheme virtual
Xmachine architecture, and present an operational definition of the instruction
Xset and the compiler.
X
X\bigskip
X
XPLScheme consists of two major parts; these are a virtual machine and a
Xcompiler.  Most of the complexity of the system lies in the implementation of
Xthe virtual machine, which implements a small instruction set tailored to the
Xexecution of Scheme.  The instruction set is in some respects similar to that
Xfound in machines designed to execute {\sc LISP}, having primitive
Xinstructions to accomplish common operations such as \code{CAR,} \code{CDR,}
Xetc.  There are dissimilarities, however.  For example, the PLScheme virtual
Xmachine implements input and output operations as single instructions.  In
Xmost implementations of {\sc LISP}- or Scheme-like languages, large subsystems
Xsuch as these are coded in the language being implemented.
X
XIn this dialect of Scheme, type-checking is performed at runtime.  In this
Xway, it resembles the Scheme dialects for which we wrote denotational and
Xoperational definitions at earlier points in the course.
X
X\section*{System Organization}
X
XAs we noted above, the PLScheme virtual machine implements an instruction set
Xtailored to the implementation of Scheme.  The storage architecture defined by
Xthis machine includes several registers that maintain the current state of the
Xcomputation; a stack, which is used to hold values of actuals being gathered
Xfor an application, as well as a chain of pending continuations; and two
Xheaps, only one of which contains accessible data while user programs are
Xbeing executed.
X
XThe virtual machine does not include the compiler, which is a separate program
Xthat generates code that can be run on the virtual machine.  As currently
Ximplemented, the PLScheme compiler is not written in PLScheme, and so cannot
Xcompile itself and thus be run in the virtual machine; instead, it is a
Xseparate program that happens to be written in Scheme, which generates code
Xthat can be run by the virtual machine.  Thus the PLScheme compiler is
Xa cross-compiler: it does not run on the machine for which it produces code.
X
XNote that, although the PLScheme virtual machine given in Handout~59 is
Ximplemented in Scheme, this is simply one implementation of a specification;
Ximplementations in other languages, such as C, or assembly code, are possible.
XSo long as the specification is adhered to, code generated by the compiler
Xwill run in the virtual machine.
X
XSee figure~\ref{sysorg} for a diagram of system organization.
X
X\begin{figure}
X\begin{verbatim}
X                           +--------------------------------------+
X                           | RUN                                  |
X                           |                                      |
X                           |   wrapped in      +---------+        |
X = PLScheme expression =>  | = (lambda ()   => | COMPILE |        |
X                           |     expression)   +---------+        |
X                           |                       ||             |
X                           |                compiled PLScheme     |
X                           |               virtual machine code   |
X                           |                       ||             |
X                           |               made into thunk with   |
X                           |                global environment    |
X                           |                       ||             |
X                           |                invocation begun      |
X                           |                       ||             |
X                           |                       \/             |
X                           |             +----------------------+ |
X                           |             |      INTERPRET       | |
X                           |             |  interprets virtual  | |
X                           |             | machine instructions | |
X                           |             +----------------------+ |
X                           +--------------------------------------+
X\end{verbatim}
X\caption{Data flow in PLScheme evaluation cycle.}
X\label{sysorg}
X\end{figure}
X
X\section*{Memory Organization}
X
XPLScheme styles memory as a contiguous vector of integers, beginning at
Xaddress 0.  As mentioned above, this is divided into three ``areas;'' these
Xare the ``current'' area, the ``other'' area, and the stack.  The current area
Xand the other area are heaps; each may contain stored objects, although only
Xone is in use at any time.\footnote{Two heaps are provided in order to allow
Xuse of a copying garbage collection algorithm due to Cheney.  The apparent
Xeffect while user code runs is that only a single heap exists; at each garbage
Xcollection, live storage is copied from the current heap to the other heap,
Xleaving garbage behind.  Then the other heap is exchanged with the current
Xheap.}
X
XIn addition to memory, several registers are maintained; these are:
X
X\begin{description}
X\item[\code{*val*}] Holds the value resulting from the most recent
Xcomputation.  It is also used to hold the first argument passed to a primitive
Xoperation; other arguments are passed on the stack.
X\item[\code{*env*}] Holds a pointer to the current environment object.
X\item[\code{*cont*}] Holds a pointer to the current continuation.
X\item[\code{*code*}] Holds a pointer to the beginning of the vector of
Xinstructions currently being executed.
X\item[\code{*pc*}] The program counter, which is an integer index into the
Xvector that is the current value of \code{*code*}.
X\item[\code{*sp*}] Holds a pointer to the value on ``top of the stack.''  The
Xstack grows from high addresses towards low addresses; thus ``popping'' the
Xstack would require dereferencing the contents of \code{*sp*} to fetch the
Xvalue currently on the stack, and then incrementing \code{*sp*;} where
X``pushing'' a value on the stack would require decrementing \code{*sp*} (so
Xthat it pointed to a fresh location on the stack) and then storing the value
Xin the location pointed to by the updated \code{*sp*.}
X\item[\code{*nargs*}] Used to pass to a procedure the number of actuals with
Xwhich it was invoked.  The \code{check-nargs} instruction checks this value on
Xprocedure entry, and a system error is signalled if it is not the same as the
Xnumber of formals with which the procedure was defined.
X\item[\code{*global-environment*}] Holds a pointer to the global environment
Xobject.
X\item[\code{*symbol-table*}] Holds a pointer to the global symbol table.  The
Xglobal symbol table contains all the symbols known to the PLScheme system,
Xand allows the reader to maintain unique symbols.
X\end{description}
X
XA diagram of the PLScheme virtual machine's storage architecture may be found
Xin figure~\ref{storage-arch}.
X
X\begin{figure}
X\begin{small}
X\begin{verbatim}
X                       Memory                                        Registers
X                   +--------------+   \                           +--------------+
Xlocation 0         |              |    |                   *val*  |              |
X                   +--------------+    |                          +--------------+
Xlocation 1         |              |    |                          +--------------+
X                   +--------------+    > first heap area   *env*  |              |
X                   |      :       |    |                          +--------------+
X                   /      :       /    |                          +--------------+
X                   /      :       /    |                   *cont* |              |
X                   |      :       |   /                           +--------------+
X                   +--------------+                               +--------------+
X                   +--------------+   \                    *code* |              |
X                   |              |    |                          +--------------+
X                   +--------------+    |                          +--------------+
X                   |              |    |                   *pc*   |              |
X                   +--------------+    > second heap area         +--------------+
X                   |      :       |    |                          +--------------+
X                   /      :       /    |                  *nargs* |              |
X                   /      :       /    |                          +--------------+
X                   |      :       |   /                           +--------------+
X                   +--------------+          *global-environment* |              |
Xlowest stack     / +--------------+                               +--------------+
Xarea location   |  |              |                               +--------------+
X     /\         |  +--------------+                *symbol-table* |              |
X     ||         |  /      :       /                               +--------------+
Xdirection of    |  +--------------+                               +--------------+
Xstack growth    |  | value on top |                       *sp*    |      **      |
X                |  |   of stack   |  <======\\                    +------||------+
Xthe stack area  <  +--------------+         ||                           ||
X                |  /      :       /         \\===========================//
X                |  +--------------+
Xhighest memory  |  |              |
Xlocation         \ +--------------+
X\end{verbatim}
X\end{small}
X\caption{The PLScheme virtual machine's storage architecture.}
X\label{storage-arch}
X\end{figure}
X
XEvery location in memory may be thought of as {\em tagged;} that is to say,
Xthere is associated with its contents a particular type, which may be one of
X
X\begin{description}
X\item[small-integer] A literal integer value of limited size, where the
Xlimits on size are implementation-dependent.
X\item[misc] A literal value that is one of \code{\#t}, \code{\#f}, the empty list,
X\code{\#\{Unspecified\}}, or \code{\#\{Undefined\}}.
X\item[pair] A pointer to an object that is a pair.
X\item[vector] A pointer to an object that is a vector.
X\item[symbol] A pointer to an object that is a symbol.
X\item[procedure] A pointer to an object that is a procedure.
X\item[ref] A pointer to an object that is an object reference.
X\end{description}
X
XAlthough we shall defer until the next lecture any details about data
Xrepresentations, it will be useful in what follows to know that primitive
Xcompound objects in PLScheme are styled as vectors, with a header word
Xspecifying the length of the vector.
X
X\section*{Virtual Machine Instruction Set}
X
XWe present an operational semantics for the PLScheme virtual machine
Xinstruction set in the form of rewrite rules, much as we did for the SECD
Xmachine of Handout 18.  The PLScheme virtual machine is considerably more
Xcomplex than the SECD machine, containing, as it does, not only a stack and a
Xheap, but also eight registers, and more instructions than the SECD machine.
XFurther differences from the SECD machine include a linear organization of
Xcode and a single stack.
X
XIn order to simplify our task, therefore, we will omit some details, and adopt
Xsome notational conventions.  In particular, we do not include in our rewrite
Xrules components of the state whose values do not change across the rule
Xtransformation.  This means that certain registers, such as
X\code{*global-environment*} and \code{*symbol-table*} are very rarely (or not at all) depicted
Xin our rewrite rules.
X
XBecause with this convention we cannot identify state elements by their
Xposition in a state tuple, we label individual elements, using two notations.
XThe notation
X\begin{center}
X\begin{tt}
Xregister = value
X\end{tt}
X\end{center}
Xmeans that \code{register} is to contain the value \code{value.} \code{value}
Xshould be understood to be an integer that may be interpreted as, say, a
Xpointer or a literal; it is what is called in the implementation a {\em
Xdescriptor.}
X
XIn this operational definition, the names of registers are given without the
Xasterisks (``{\tt *}'') that by custom surround the names of Scheme global
Xvariables.  Thus the environment register \code{*env*,} for example, is shown
Xsimply as \code{env.}
X
XWe may also write
X\begin{center}
X\begin{tt}
X@register = value
X\end{tt}
X\end{center}
Xin which case we understand \code{value} to be an object in the heap or on the
Xstack, or possibly a representation of some range of the stack itself; in this
Xcase we mean that \code{register} is to contain a pointer to \code{value.} Put
Xanother way, ``\code{@}'' signifies indirection.  For example, we may write
X\begin{center}
X\begin{tt}
X@sp = <v1 v2 . s>
X\end{tt}
X\end{center}
Xto mean that the stack pointer is pointing to a stack slot holding the value
X\code{v1,} and that immediately below it on the stack (i.e., at the next higher
Xaddress) is the value \code{v2,} and that the remainder of the stack may be
Xrepresented as simply \code{s} (note that ``.'' here is used on vectors in the
Xsame way that it was used on lists in Handout 18).
X
XOur notation represents environments as vectors containing tuples of values.
XIn the following:
X\begin{center}
X\begin{tt}
X@env = <<v$_{01}$ \ldots v$_{0n_0}$> \ldots <v$_{i1}$ \ldots v$_{ij}$ \ldots v$_{in_i}$> . e>
X\end{tt}
X\end{center}
Xwe are saying that the contents of the environment register is a pointer to an
Xenvironment in which lexical variable $\pair{i}{j}$ has value ${\tt v}_{ij}$.
X
XBecause we can do so unambiguously, we take some liberties with our depiction
Xof the program counter.  We say
X\begin{center}
X\begin{tt}
X@code[pc] = <load-variable i j . c>
X\end{tt}
X\end{center}
Xto mean that the register \code{*code*} offset by the register \code{*pc*}
Xpoints to a portion of code whose first element is the instruction
X\code{load-variable,} whose second element is \code{i}, and so on.  \code{c}
Xrepresents the code following the particular sequence of interest.  We may use
Xthis notation unambiguously, because code vectors never overlap; thus only one
Xpair of values for the registers \code{*code*} and \code{*pc*} may point to a
Xparticular instruction.
X
XWhen we wish to describe a descriptor in detail, giving its particular tag, we
Xuse the format \code{[{\em tag,} value]}.  If \code{value} is a compound
Xstructure of some sort, it is understood that a pointer to \code{value} is
Xintended.  When we wish to describe a vector in the heap or on the stack, we
Xsimply list its elements, using the notation \code{\#H(e$_1$ \ldots e$_n$)} to
Xdenote a vector in the heap, and \code{\#S(e$_1$ \ldots e$_n$)} to denote a
Xstack vector.  Thus in particular, we may use the notation \code{[{\em tag},
X\#H(e$_1$ \ldots e$_n$)]} to denote a pointer to a heap-allocated vector
Xwith $n$\/ elements, where the pointer has the tag {\em tag.}
X
X\subsection*{Rewrite Rules}
X
X\bigskip
X
X\noindent \code{(check-nargs i)}: Check that number of args is i
X\begin{description}
X\item[\rulename{check-nargs}]
X\[{\code{@code[pc] = <check-nargs i . c>, nargs = i}}
X\over
X{\code{@code[pc] = c, nargs = i}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(make-environment n)}: Make an environment object from the top
Xn values on the stack
X\begin{description}
X\item[\rulename{make-environment}]
X\[{\code{@code[pc] = <make-environment n . c>, env = e, @sp = <v$_1$ \ldots v$_n$ . s>}}
X\over
X{\code{@code[pc] = c, @env = <<v$_1$ \ldots v$_n$> . e>, @sp = s}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(load-constant x)}: Load next word into value register
X\begin{description}
X\item[\rulename{load-constant}]
X\[{\code{@code[pc] = <load-constant x . c>}}
X\over
X{\code{@code[pc] = c, val = x}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(load-variable i j)}: Load the variable at lexical offset
X$\pair{i}{j}$ into the value register
X\begin{description}
X\item[\rulename{load-variable}]
X\[{\begin{array}{c}
X  \code{@code[pc] = <load-variable i j . c>,}\\
X \code{@env = <<v$_{01}$ \ldots v$_{0n_0}$> \ldots <v$_{i1}$ \ldots
X         v$_{ij}$ \ldots v$_{in_i}$> . e>}
X\end{array}}
X\over
X{\begin{array}{c}
X\code{@code[pc] = c, val = v$_{ij}$,}
X       \\ \code{@env = <<v$_{01}$ \ldots v$_{0n_0}$> \ldots <v$_{i1}$ \ldots
X                         v$_{ij}$ \ldots v$_{in_i}$> . e>}
X\end{array}}\] 
X\end{description}
X
X\bigskip
X
X\noindent \code{(set-variable i j)}: Store the value register contents in the
Xvariable at lexical offset $\pair{i}{j}$.
X\begin{description}
X\item[\rulename{set-variable}]
X\[{\begin{array}{c}
X\code{@code[pc] = <set-variable i j . c>, val = v,}\\
X   \code{@env = <<v$_{01}$ \ldots v$_{0n_0}$> \ldots <v$_{i1}$ \ldots
X        v$_{ij-1}$, v$_{ij}$, v$_{ij+1}$ \ldots v$_{in_i}$> . e>}
X\end{array}}
X\over
X{\begin{array}{c}
X\code{@code[pc] = c, val = v,}
X       \\ \code{@env = <<v$_{01}$ \ldots v$_{0n_0}$> \ldots <v$_{i1}$ \ldots
X                         v$_{ij-1}$, v, v$_{ij+1}$ \ldots v$_{in_i}$> . e>}
X\end{array}}\] 
X\end{description}
X
X\bigskip
X
X\noindent \code{(make-procedure o)}: Make procedure object, using next word
Xfor code
X\begin{description}
X\item[\rulename{make-procedure}]
X\[{\code{@code[pc] = <make-procedure o . c>, env = e}}
X\over
X{\code{@code[pc] = c, val = [procedure, \#H(o e)]}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(push)}: Push the value in the value register on the stack
X\begin{description}
X\item[\rulename{push}]
X\[{\code{@code[pc] = <push . c>, val = v, @sp = s}}
X\over
X{\code{@code[pc] = c, val = v, @sp = <v . s>}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(call n)}: invoke a procedure with n arguments
X\begin{description}
X\item[\rulename{call}]
X\[{\code{@code[pc] = <call n . c>, val = [procedure, \#H(o e)]}}
X\over
X{\code{code = o, pc = 0, nargs = n, env = e, val = [procedure, \#H(o e)]}}\]
X\end{description}
X
X\bigskip
X% Error c -> k pointed out by David Beymer or Tao Alter
X\noindent \code{(make-continuation p i)}: Create a continuation on the stack
X\begin{description}
X\item[\rulename{make-continuation}]
X\[{\begin{array}{c}
X\code{@code[pc] = <make-continuation p i . c>, code = o,}
X         \\ \code{env = e, cont = k, @sp = <v$_1$ \ldots v$_i$ . s>}
X\end{array}}
X\over
X{\begin{array}{c}
X\code{@code[pc] = c, env = e, code = o,
X       cont = [vector, sp],}\\
X       \code{@sp = <(i + 4) k o [small-integer, p] e v$_1$ \ldots v$_i$ . s>}
X\end{array}}\]
X\end{description}
X
X\noindent The argument \code{i} to \code{make-continuation} is the number
Xof values on top of the stack that are not currently a part of any
Xstack-allocated stored object.  When making the continuation, for consistency's
Xsake we'd like to store these values in the continuation object (i.e. a stack-allocated vector) so that
Xthey belong to {\em some} stored object on the stack.  Thus the size of the
Xcontinuation vector (excluding header word) will be \code{i} (the number of
X``unstored'' values on top of the stack) $+~4$ (the number of register values
Xwe push on the stack to form a continuation).  After the \code{make-continuation}
Xis performed, the values on the top of the stack,
X
X\centerobj{\code{(i + 4) c o [small-integer, p] e v$_1$ \ldots v$_i$}}
X
X\noindent actually comprise a stack vector that could be
Xwritten with the notation \code{\#S(...)}.
X
X\bigskip
X
X\noindent \code{(return)}: Resume the continuation in the continuation register.
X(Here \code{k} refers to the number of values on the stack that are ``above''
Xthe current continuation; these are automatically deallocated by \code{return})
X
X\begin{description}
X\item[\rulename{return}]
X\[{\begin{array}{c}
X\code{@code[pc] = <return . c>,
X       cont = [vector, sp + k],}\\
X       \code{@sp = <x$_1$ \ldots x$_k$, i + 4, c o [small-integer, p] e v$_1$
X\ldots v$_i$ . s>}
X\end{array}}
X\over
X{\code{code = o, pc = p, env = e, cont = c, @sp = <v$_1$ \ldots v$_i$ . s>}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(leave-environment)}: Leave a lexical contour
X\begin{description}
X\item[\rulename{leave-environment}]
X\[{\code{@code[pc] = <leave-environment . c>, @env = <<v$_1$ \ldots v$_n$> . e>}}
X\over
X{\code{@code[pc] = c, @env = e}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(jump-if-false p)}: Branch to offset on value register false
X\begin{description}
X\item[\rulename{jump-if-false {\rm (fails)}}]
X\[{\code{@code[pc] = <jump-if-false p . c>, val = \#t}}
X\over
X{\code{@code[pc] = c, val = \#t}}\]
X\item[\rulename{jump-if-false {\rm (succeeds)}}]
X\[{\code{@code[pc] = <jump-if-false p . c>, val = \#f}}
X\over
X{\code{pc = p, val = \#f}}\]
X\end{description}
X
X\bigskip
X
X\noindent \code{(jump p)}: Branch to offset always
X\begin{description}
X\item[\rulename{jump}]
X\[{\code{@code[pc] = <jump p . c>}}
X\over
X{\code{pc = p}}\]
X\end{description}
X
X\bigskip
X
X\noindent {\em Primitive Instructions.} In addition to the above instructions, 
Xthe PLScheme instruction set includes a wide variety of instructions for
Xoperating on numbers, pairs, vectors, and refs and for performing input/output.
XAlthough individual instructions perform different data manipulations,
Xtheir rewrite rules all look alike. Unlike
Xuser-defined procedures, which expect all arguments on top of the stack, 
Xthe primitive instructions expect their first argument in the \code{*val*} register,
Xand all arguments but the first on the stack.  The following rewrite rules for
Xthe primitive addition operation exemplifies  how the virtual machine
Xhandles these primitive instructions:
X
X\smallskip
X
X\noindent \code{(+)}: Add the number in the \code{*val*} register to the 
Xnumber at the top of the stack.
X\begin{description}
X\item[\rulename{+}]
X\[{\code{@code[pc] = <+ . c>, val = [small-integer, $\fun{num}_1$], \\
X                     @sp = <[small-integer, $\fun{num}_2$] . s>}}
X\over
X{\code{@code[pc] = c, val = [small-integer, $\fun{num}_1 + \fun{num}_2$], @sp = s}}\]
X\end{description}
X
X\newpage
X
X\section*{Compiler Rules}
X
X\newcommand{\boxit}[1]{\begin{center}
X\fbox{
X\begin{minipage}[t]{5.75in}
X{#1}
X\end{minipage}}
X\end{center}
X}
X
X\newcommand{\compdots}{$~~~~\vdots$\\}
X\newcommand{\largecode}[1]{\mbox{\large\tt #1}}
X\newcommand{\compenv}{{\em env}}
X\newcommand{\compnext}{{\em NEXT}}
X\newcommand{\compreturn}{{\em RETURN}}
X\newcommand{\compjump}[1]{\code{\em (JUMP #1)}}
X\newcommand{\company}{{\em cont}}
X\newcommand{\comprest}{<rest of instructions>}
X\newcommand{\complabel}[1]{{\em #1}}
X\newcommand{\compbind}[1]{$\fun{bind}($#1, \compenv$)$}
X\newcommand{\comptitle}[1]{\noindent {\large\bf #1}}
X\newcommand{\compcont}[1]{
X
X\noindent {$\cal K$} $($#1$)$
X
X}
X\newcommand{\compinstr}[1]{
X
X\noindent ~~\largecode{#1}
X
X}
X\newcommand{\comp}[3]{\vspace{0.05in}
X
X\noindent \mbox{\hbox{$\cal C$}[~\code{#1}, ~#2, ~#3~]}
X
X\vspace{0.05in}}
X
X\newcommand{\comprule}[4]{\noindent\boxit{\comp{#1}{#2}{#3}
X\vspace{0.05in}
X\noindent $~~~~~~~~~~~~~~~~~~~\Downarrow$\\
X\noindent #4}
X}
X
X\newcommand{\compcomment}[4]{\vspace{0.05in}
X
X\noindent \mbox{\hbox{$\cal C$}[~\code{#1}, ~#2, ~#3~] #4}
X
X\vspace{0.05in}}
X
X\newcommand{\compcommentrule}[5]{\noindent\boxit{\compcomment{#1}{#2}{#3}{\code{#4}}
X\vspace{0.05in}
X\noindent $~~~~~~~~~~~~~~~~~~~\Downarrow$\\
X\noindent #5}
X}
X
X\newcommand{\compindent}[4]{\vspace{0.05in}
X
X\noindent {#1}\mbox{\hbox{$\cal C$}[~\code{#2}, ~#3, ~#4~]}
X
X\vspace{0.05in}}
X
X\newcommand{\krule}[2]{\noindent {$\cal K$} $($#1$)$ emits {#2}\\}
X
X
X
XPLScheme programs are compiled into the opcodes of the 
XPLScheme virtual machine by a simple compiler.  The 
Xflow of information through the stages of the compiler
Xis outlined in the diagram in figure~\ref{compiler}.
X
XThe compiler begins with a no-argument lambda expression, which is first
Xdesugared into a simple subset of PLScheme.  The desugaring phase
Xtranslates (in the standard way) such special forms as \code{let}, 
X\code{and}, etc. into other forms of the language.  The simplified
Xexpression is then compiled into an assembly code that makes use of
Xsymbolic opcodes and symbolic labels.  Finally, the resulting assembly
Xcode is assembled into the low-level object code for 
Xthe virtual machine.
X
XBelow we describe how assembly code for the desugared subset of
XPLScheme is generated.  We will provide a set of compiler rules in a
Xmanner similar to the one we used for the SECD machine in Handout~18.
XHere we use the notation 
X
X\begin{center}
X\comp{{\em exp}}{\compenv}{\company}
X\end{center}
X
X\noindent to designate the sequence of instructions generated by compiling the
Xdesugared PLScheme expression {\em exp} in the compile-time environment
X{\em env} with the continuation {\em cont}.  
X
XThe continuation used by the compiler is closely related to the
Xcontinuations we saw in standard semantics.  Just as the valuation
Xfunction for an expression uses an expression continuation to specify
Xwhat to do with the result of the expression, the compiler uses 
Xits continuation to determine ``who'' is going to consume the value
Xproduced when the compiled code for a given expression is executed.
XThere are three different kinds of continuations used by the PLScheme
Xcompiler:
X
X\begin{enumerate}
X
X\item \compnext : Indicates that the value will be consumed by the
Ximmediately following code sequence.
X
X\item \compreturn : Indicates that the value will be consumed by
Xwhatever piece of code the the currently executing procedure
Xreturns to.
X
X\item \compjump{label} : Indicates that the value will be 
Xconsumed by the code sequence beginning at the label \complabel{label}.
X
X\end{enumerate}
X
X\noindent When the type of continuation does not affect the code
Xtemplates below, the continuation will be specified as \company,
Xwhich stands for any of the above three continuations (unless
Xsome further restriction is specified).
X
XJust as a standard semantics continuation is only invoked when
Xan actual value is produced, the compiler continuation only generates
Xcode (if at all) when a primitive value-producing expression is compiled.
XThe code generated in these cases is denoted  $\cal K(\fun{cont})$ below.
X
XThe compiler rules presented below are simplified slightly in that
Xno reference is made to the {\em depth} parameter that is an 
Xadditional parameter to the actual compilation procedures.  The only 
Xpurpose of this parameter is to help the compiler determine
Xhow big a stack-vector will be needed when a \code{make-continuation} instruction
Xis performed.  Recall that the \code{make-continuation} instruction 
Xneeds to know how many ``unstored'' values are on top of the stack
Xso that they can be stored in the same stack-allocated vector as
Xthe continuation state.  The compiler uses the {\em depth} parameter
Xto count how many unaccounted for values have been pushed on the
Xstack.  Since this is largely an uninteresting technical point,
Xthe {\em depth} parameter has been left out of the compiler rules.
X
XThere are several situations in the rules below where a 
Xnew label (such as {\em return-label}, {\em else-label},
Xor {\em after-if}~ label) is introduced.  In these cases 
Xyou are to assume that a unique label is generated by the
Xcompiler; this label will be thus be different for different applications
Xof the rule.
X
X\begin{figure}[h]
X\begin{verbatim}
X                      Lambda Expression
X                            ||
X             +--------------||------------------+
X             | COMPILE      ||                  |
X             |              ||                  |
X             |              \/                  |
X             |        +-------------+           |
X             |        |  DESUGARER  |           |
X             |        +-------------+           |
X             |              ||                  |
X             |   Desugared PLScheme program     |
X             |              ||                  |
X             |              \/                  |
X             |      +------------------+        |
X             |      |  CODE GENERATOR  |        |
X             |      +------------------+        |
X             |              ||                  |
X             |       Assembly code for          |
X             |        virtual machine           |
X             |              ||                  |
X             |              \/                  |
X             |        +-------------+           |
X             |        |  ASSEMBLER  |           |
X             |        +-------------+           |
X             |              ||                  |
X             +--------------||------------------+
X                      Object code for 
X                      virtual machine
X                            ||
X                            \/
X\end{verbatim}
X\caption{Stages of the PLScheme compiler}
X\label{compiler}
X\end{figure}
X
X\newpage
X
X\comptitle{Basic Continuations}
X
X\boxit{
X\krule{\compnext}{no code}
X\krule{\compreturn}{\largecode{(RETURN)}}
X\krule{\compjump{label}}{\largecode{(JUMP \complabel{label})}}
X}
X
X\comptitle{Literal Expressions}
X
X\comprule{$\fun{literal}$}{\compenv}{\company}
X{\compinstr{(LOAD-CONSTANT (LITERAL $\fun{literal}$))}
X \compcont{\company}
X }
X
X\comptitle{Variable References}
X
X\comprule{$\fun{primitive-ref}$}{\compenv}{\company}
X{\comp{(lambda ($v_1$ $\ldots$ $v_n$) ($\fun{primitive-ref}$ $v_1$ $\ldots$ $v_n$))}{\compenv}{\company}
X }
X
X\noindent The above rule is only invoked when a Scheme primitive operator
Xname occurs in some position other than the operator position of 
Xa combination.  In this case, an eta-expansion is used to wrap
Xthe operator within a procedure so that it can be treated as 
Xa first-class procedure object.  The case where a Scheme primitive
Xoperator name occurs in the operator position of a combination 
Xis handled by the {\bf Open Procedure Calls} rule (below).
X
X\comprule{$\fun{variable-ref}$}{\compenv}{\company}
X{\compinstr{(VARIABLE {\sc lex-depth({\em variable-ref}, {\em env})}
X                      {\sc lex-offset({\em variable-ref}, {\em env})})}
X \compcont{\company}
X }
X
X\noindent Here, {\sc lex-depth} and {\sc lex-offset} are auxiliary
Xfunctions that compute the relative position (in terms of lexical depth and
Xlexical offset) of {\em variable-ref} in the
Xcurrent compile-time environment {\em env}.
X
X\newpage
X
X\comptitle{BEGIN Expressions}
X
X\comprule{(begin $e_1$ $e_2$ $\ldots$ $e_n$)}{\compenv}{\company}
X{\comp{$e_1$}{\compenv}{\compnext}
X \comp{$e_2$}{\compenv}{\compnext}
X \compdots
X \comp{$e_n$}{\compenv}{\company}
X }
X
X
X\comptitle{IF Expressions}
X
X\comprule{(if $e_1$ $e_2$ $e_3$)}{\compenv}{\compnext}
X{\comp{$e_1$}{\compenv}{\compnext}
X \compinstr{(JUMP-IF-FALSE \complabel{else-label})}
X \comp{$e_2$}{\compenv}{\compjump{after-if-label}}
X \compinstr{\complabel{else-label}}
X \comp{$e_3$}{\compenv}{\compnext}
X \compinstr{\complabel{after-if-label}}
X }
X
X\compcommentrule{(if $e_1$ $e_2$ $e_3$)}{\compenv}{\company}{    ; \company $~\neq$ \compnext}
X{\comp{$e_1$}{\compenv}{\compnext}
X \compinstr{(JUMP-IF-FALSE \complabel{else-label})}
X \comp{$e_2$}{\compenv}{\company}
X \compinstr{\complabel{else-label}}
X \comp{$e_3$}{\compenv}{\company}
X }
X
X\newpage
X
X\comptitle{LAMBDA Expressions}
X
X\comprule{(lambda ($v_1$ $\ldots$ $v_n$) $e$)}{\compenv}{\company}
X{\compinstr{(MAKE-PROCEDURE}
X \compinstr{~~(CODE}
X \compinstr{~~~~(CHECK-NARGS $n$)}
X \compinstr{~~~~(MAKE-ENVIRONMENT $n$)}
X \compindent{~~~~~~~~}{$e$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compreturn}
X \compcont{\company}
X }
X
X\comptitle{Open Procedure Calls}
X
X\comprule{($\fun{Scheme-primop}$ $e_1$ $e_2$ $\ldots$ $e_n$)}{\compenv}{\company}
X{\comp{$e_n$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \compdots
X \comp{$e_2$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \comp{$e_1$}{\compenv}{\compnext}
X \compinstr{{\sc virtual-machine-primitive($\fun{Scheme-primop}$)}}
X \compcont{\company}
X }
X
X\noindent where {\sc virtual-machine-primitive} is an auxiliary function
Xthat returns the machine instruction corresponding to the Scheme primitive
Xoperator.
X
X\newpage
X
X\comptitle{Closed Procedure Calls}
X
X\comprule{($e_0$ $e_1$ $\ldots$ $e_n$)}{\compenv}{\compnext}
X{\compinstr{(MAKE-CONTINUATION \complabel{return-label})}
X \comp{$e_n$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \compdots
X \comp{$e_1$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \comp{$e_0$}{\compenv}{\compnext}
X \compinstr{(CALL $n$)}
X \compinstr{\complabel{return-label}}
X }
X
X\comprule{($e_0$ $e_1$ $\ldots$ $e_n$)}{\compenv}{\compjump{jump-label}}
X{\compinstr{(MAKE-CONTINUATION \complabel{jump-label})}
X \comp{$e_n$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \compdots
X \comp{$e_1$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \comp{$e_0$}{\compenv}{\compnext}
X \compinstr{(CALL $n$)}
X }
X
X\comprule{($e_0$ $e_1$ $\ldots$ $e_n$)}{\compenv}{\compreturn}
X{\comp{$e_n$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \compdots
X \comp{$e_1$}{\compenv}{\compnext}
X \compinstr{(PUSH)}
X \comp{$e_0$}{\compenv}{\compnext}
X \compinstr{(CALL $n$)}
X }
X 
X\newpage
X
X\comptitle{LETREC Expressions}
X
X\comprule{(letrec (($v_1$ $e_1$) $\ldots$ ($v_n$ $e_n$)) $e$)}{\compenv}{\compreturn}
X{\compinstr{(LOAD-CONSTANT UNSPECIFIED)}
X \compinstr{(PUSH)}
X \compdots
X \compinstr{{\em ; Perform the LOAD/PUSH pair of instructions N times in all}}
X \compinstr{(LOAD-CONSTANT UNSPECIFIED)}
X \compinstr{(PUSH)}
X \compinstr{(MAKE-ENVIRONMENT $n$)}
X \comp{$e_n$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compnext}
X \compinstr{(SET-VARIABLE 0 $n$)}
X \compdots
X \comp{$e_1$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compnext}
X \compinstr{(SET-VARIABLE 0 1)}
X \comp{$e$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compreturn}
X }
X
X\compcommentrule{(letrec (($v_1$ $e_1$) $\ldots$ ($v_n$ $e_n$)) $e$)}{\compenv}{\company}{    ; \company $~\neq$ \compreturn}
X{\compinstr{(LOAD-CONSTANT UNSPECIFIED)}
X \compinstr{(PUSH)}
X \compdots
X \compinstr{{\em ; Perform the LOAD/PUSH pair of instructions N times in all}}
X \compinstr{(LOAD-CONSTANT UNSPECIFIED)}
X \compinstr{(PUSH)}
X \compinstr{(MAKE-ENVIRONMENT $n$)}
X \comp{$e_n$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compnext}
X \compinstr{(SET-VARIABLE 0 $n$)}
X \compdots
X \comp{$e_1$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compnext}
X \compinstr{(SET-VARIABLE 0 1)}
X \comp{$e$}{\compbind{\code{($v_1$ $\ldots$ $v_n$)}}}{\compnext}
X \compinstr{(LEAVE-ENVIRONMENT)}
X \compcont{\company} % Pointed out by Tao Alter
X }
X
X\end{body}
X
X
X
X
X
X
X
X
X
X
X
X
X
X
X
X
X
END_OF_FILE
if test 35366 -ne `wc -c <'plscheme/doc/plscheme-1.tex'`; then
    echo shar: \"'plscheme/doc/plscheme-1.tex'\" unpacked with wrong size!
fi
# end of 'plscheme/doc/plscheme-1.tex'
fi
if test -f 'plscheme/vm.scm' -a "${1}" != "-c" ; then 
  echo shar: Will not clobber existing file \"'plscheme/vm.scm'\"
else
echo shar: Extracting \"'plscheme/vm.scm'\" \(28360 characters\)
sed "s/^X//" >'plscheme/vm.scm' <<'END_OF_FILE'
X;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
X;;;                 This is the file VM.SCM                        ;;;
X;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
X
X;;; VIRTUAL MACHINE: DATA REPRESENTATION AND MEMORY LAYOUT
X
X; Handy utility for internal consistency checking
X
X(define (assert truth where)
X  (if (not truth)
X      (error "assertion failed" where)))
X
X; Statistics counters
X
X(define *counters* '())
X
X(define (make-counter name)
X  (let ((counter (vector 0)))
X    (set! *counters* (cons (list name counter) *counters*))
X    counter))
X
X(define (add-to-counter counter how-much)
X  (vector-set! counter 0 (+ (vector-ref counter 0) how-much)))
X
X(define (reset-counters)
X  (for-each (lambda (name+counter)
X              (vector-set! (cadr name+counter) 0 0))
X            *counters*))
X
X(define (show-counters)
X  (for-each (lambda (name+counter)
X              (newline)
X              (display (car name+counter))
X              (display " ")
X              (write (vector-ref (cadr name+counter) 0)))
X            *counters*)
X  (newline))
X
X; Registers
X
X(define (make-register initial-contents)
X  (vector 'register initial-contents))
X
X(define (fetch register)
X  (vector-ref register 1))
X
X(define (store! register new-value)
X  (vector-set! register 1 new-value))
X
X(define *descriptor-registers* '())
X
X(define (make-descriptor-register)
X  (let ((register (make-register vm-unspecified)))
X    (set! *descriptor-registers*
X          (cons register *descriptor-registers*))
X    register))
X
X(define (clear-descriptor-registers)
X  (for-each (lambda (reg) (store! reg vm-unspecified))
X            *descriptor-registers*))
X
X; Descriptors
X;  A descriptor describes a Scheme object.
X
X(define tag-limit 8)
X
X(define (make-descriptor tag data)
X  (assert (and (integer? data) (integer? tag) (>= tag 0) (< tag tag-limit))
X          'make-descriptor)
X  (+ tag (* data tag-limit)))
X
X(define (descriptor-tag descriptor) (remainder descriptor tag-limit))
X(define (descriptor-data descriptor) (quotient descriptor tag-limit))
X
X(define same-descriptor? =)
X
X; Tags
X
X(define small-integer-tag 0)
X(define misc-tag          1)
X; tag 2 is reserved for future expansion
X(define pair-tag          3)
X(define vector-tag          4)
X(define symbol-tag          5)
X(define procedure-tag          6)
X(define ref-tag                  7)
X
X(define least-stored-object-tag pair-tag)
X
X(define (stored-object? descriptor)
X  (>= (descriptor-tag descriptor) least-stored-object-tag))
X
X(define (vm-small-integer? descriptor)
X  (= (descriptor-tag descriptor) small-integer-tag))
X
X(define (vm-pair? descriptor)
X  (= (descriptor-tag descriptor) pair-tag))
X
X(define (vm-vector? descriptor)
X  (= (descriptor-tag descriptor) vector-tag))
X
X(define (vm-symbol? descriptor)
X  (= (descriptor-tag descriptor) symbol-tag))
X
X(define (vm-procedure? descriptor)
X  (= (descriptor-tag descriptor) procedure-tag))
X
X(define (vm-ref? descriptor)
X  (= (descriptor-tag descriptor) ref-tag))
X
X; Small integers
X
X(define (vm-make-small-integer integer)
X  (make-descriptor small-integer-tag integer))
X
X(define (vm-small-integer-value descriptor)
X  (assert (vm-small-integer? descriptor) 'vm-small-integer-value)
X  (descriptor-data descriptor))
X
X; Miscellaneous objects: booleans, empty list, unspecified
X
X(define vm-false       (make-descriptor misc-tag 0))
X(define vm-true               (make-descriptor misc-tag 1))
X(define vm-empty-list  (make-descriptor misc-tag 2))
X(define vm-unspecified (make-descriptor misc-tag 3))
X(define vm-undefined   (make-descriptor misc-tag 4))
X
X(define (vm-make-boolean boolean)
X  (if boolean vm-true vm-false))
X
X(define (vm-boolean? descriptor)
X  (or (same-descriptor? descriptor vm-false)
X      (same-descriptor? descriptor vm-true)))
X
X(define (vm-boolean-value descriptor)
X  (assert (vm-boolean? descriptor) 'vm-boolean-value)
X  (not (same-descriptor? descriptor vm-false)))
X
X(define (vm-empty-list? descriptor)
X  (same-descriptor? descriptor vm-empty-list))
X
X; Descriptors for stored objects
X
X(define (make-stored-object-descriptor tag length-address)
X  (assert (>= tag least-stored-object-tag) 'make-stored-object-descriptor)
X  (make-descriptor tag length-address))
X
X(define (address-of-length-cell descriptor)
X  (assert (stored-object? descriptor) 'address-of-length-cell)
X  (descriptor-data descriptor))
X
X; (No uses of DESCRIPTOR-DATA should follow this point.)
X
X
X
X; Memory is a vector of cells.  Each cell can hold one descriptor.
X
X(define *memory* #f)
X(define *heap-size* 5000)
X(define *stack-size* 5000)
X
X(define (initialize-memory size)
X  (set! *memory* (make-vector size)))
X
X; Memory is subdivided into "areas".
X
X(define (make-area start size) (vector 'area start (+ start size) start))
X(define (area-begin area) (vector-ref area 1))
X(define (area-end area) (vector-ref area 2))
X(define (next-available area) (vector-ref area 3))
X(define (set-next-available! area new-value)
X  (vector-set! area 3 new-value))
X
X(define (in-area? descriptor area)
X  (assert (stored-object? descriptor) 'in-area?)
X  (let ((address (descriptor-data descriptor)))
X    (and (>= address (area-begin area))
X         (< address (area-end area)))))
X
X(define *current-area* #f)
X(define *other-area* #f)
X(define *stack-area* #f)
X
X(define (initialize-areas)
X  (set! *current-area* (make-area 0 *heap-size*))
X  (set! *other-area*   (make-area (area-end *current-area*) *heap-size*))
X  (set! *stack-area*   (make-area (area-end *other-area*) *stack-size*)))
X
X; Stored objects
X
X(define (length-or-forwarded-to descriptor)
X  (assert (stored-object? descriptor) 'length-or-forwarded-to)
X  (vector-ref *memory* (address-of-length-cell descriptor)))
X
X(define (set-length-or-forwarded-to! descriptor new)
X  (assert (stored-object? descriptor) 'set-length-or-forwarded-to!)
X  (vector-set! *memory* (address-of-length-cell descriptor) new))
X
X(define (forwarded? descriptor)
X  (not (vm-small-integer? (length-or-forwarded-to descriptor))))
X
X(define forwarded-to length-or-forwarded-to)
X
X(define set-forwarded-to! set-length-or-forwarded-to!)
X
X(define (stored-object-length descriptor)
X  (assert (not (forwarded? descriptor)) 'stored-object-length)
X  (vm-small-integer-value (length-or-forwarded-to descriptor)))
X
X(define (stored-object-ref descriptor offset)
X  (assert (and (stored-object? descriptor)
X               (>= offset 0)
X               (< offset (stored-object-length descriptor)))
X          'stored-object-ref)
X  (vector-ref *memory* (+ (address-of-length-cell descriptor) 1 offset)))
X
X(define (stored-object-set! descriptor offset new-component)
X  (assert (and (stored-object? descriptor)
X               (>= offset 0)
X               (< offset (stored-object-length descriptor)))
X          'stored-object-set!)
X  (vector-set! *memory*
X               (+ (address-of-length-cell descriptor) 1 offset)
X               new-component))
X
X(define heap-allocation-counter (make-counter "Heap allocated"))
X
X(define (allocate-stored-object tag size)
X  (assert (>= size 0)
X          'allocate-stored-object)
X  (cond ((> (+ (next-available *current-area*) size 1)
X            (area-end *current-area*))
X         (collect)
X         (if (> (+ (next-available *current-area*) size 1)
X                (area-end *current-area*))
X             (error "out of memory" tag size))))
X  (let ((descriptor
X         (make-stored-object-descriptor tag (next-available *current-area*))))
X    (set-length-or-forwarded-to! descriptor (vm-make-small-integer size))
X    (set-next-available! *current-area*
X                         (+ (next-available *current-area*) size 1))
X    (add-to-counter heap-allocation-counter (+ size 1))
X    descriptor))
X
X; Pairs
X
X(define (vm-make-pair)
X  (allocate-stored-object pair-tag 2))
X
X(define (vm-car descriptor)
X  (assert (vm-pair? descriptor) 'vm-car)
X  (stored-object-ref descriptor 0))
X
X(define (vm-cdr descriptor)
X  (assert (vm-pair? descriptor) 'vm-cdr)
X  (stored-object-ref descriptor 1))
X
X(define (vm-set-car! descriptor new-car)
X  (assert (vm-pair? descriptor) 'vm-set-car!)
X  (stored-object-set! descriptor 0 new-car))
X
X(define (vm-set-cdr! descriptor new-cdr)
X  (assert (vm-pair? descriptor) 'vm-set-cdr!)
X  (stored-object-set! descriptor 1 new-cdr))
X
X; Refs are just like pairs except that they only have one component.
X
X(define (vm-make-ref)
X  (allocate-stored-object ref-tag 1))
X
X(define (vm-get descriptor)
X  (assert (vm-ref? descriptor) 'vm-get)
X  (stored-object-ref descriptor 0))
X
X(define (vm-put! descriptor new-get)
X  (assert (vm-ref? descriptor) 'vm-put!)
X  (stored-object-set! descriptor 0 new-get))
X
X; Vectors
X
X(define (vm-make-vector size)
X  (allocate-stored-object vector-tag size))
X
X(define (vm-vector-length descriptor)
X  (assert (vm-vector? descriptor) 'vm-vector-length)
X  (stored-object-length descriptor))
X
X(define (vm-vector-ref descriptor index)
X  (assert (vm-vector? descriptor) 'vm-vector-ref)
X  (stored-object-ref descriptor index))
X
X(define (vm-vector-set! descriptor index new-component)
X  (assert (vm-vector? descriptor) 'vm-vector-set!)
X  (stored-object-set! descriptor index new-component))
X
X(define (vm-vector-fill! descriptor initial-value)
X  (let ((len (vm-vector-length descriptor)))
X    (do ((i 0 (+ i 1)))
X        ((>= i len))
X      (vm-vector-set! descriptor i initial-value))))
X
X; Symbols
X
X(define *symbol-table* (make-descriptor-register))
X
X(define (reset-symbol-table)
X  (store! *symbol-table* vm-empty-list))
X
X(define (find-or-make-symbol symbol)
X  (let loop ((l (fetch *symbol-table*)))
X    (cond ((vm-empty-list? l)
X           (push (vm-make-pair))
X           (let ((new-symbol (copy-symbol-to-heap symbol)))
X             (let ((pair (pop)))
X               (vm-set-car! pair new-symbol)
X               (vm-set-cdr! pair (fetch *symbol-table*))
X               (store! *symbol-table* pair)
X               new-symbol)))
X          ((same-as-vm-symbol? symbol (vm-car l))
X           (vm-car l))
X          (else
X           (loop (vm-cdr l))))))
X
X(define (same-as-vm-symbol? symbol descriptor)
X  (let ((name (symbol->string symbol)))
X    (let ((len (string-length name)))
X      (and (= len (stored-object-length descriptor))
X           (let loop ((i 0))
X             (cond ((>= i len) #t)
X                   ((not (= (vm-small-integer-value
X                              (stored-object-ref descriptor i))
X                            (char->integer (string-ref name i))))
X                    #f)
X                   (else (loop (+ i 1)))))))))
X
X(define (copy-symbol-to-heap symbol)
X  (let ((name (symbol->string symbol)))
X    (let ((len (string-length name)))
X      (let ((descriptor (allocate-stored-object symbol-tag len)))
X        (do ((i 0 (+ i 1)))
X            ((>= i len) descriptor)
X          (stored-object-set! descriptor
X                              i
X                              (vm-make-small-integer
X                                (char->integer (string-ref name i)))))))))
X
X(define (extract-symbol-from-heap descriptor)
X  (let ((len (stored-object-length descriptor)))
X    (do ((i (- len 1) (- i 1))
X         (l '() (cons (integer->char (vm-small-integer-value
X                                       (stored-object-ref descriptor i)))
X                      l)))
X        ((< i 0) (string->symbol (list->string l))))))
X
X; Procedures
X;  Procedures are represented as (code, environment) pairs.
X
X(define (vm-procedure-code descriptor)
X  (assert (vm-procedure? descriptor) 'vm-procedure-code)
X  (stored-object-ref descriptor 0))
X
X(define (vm-procedure-environment descriptor)
X  (assert (vm-procedure? descriptor) 'vm-procedure-environment)
X  (stored-object-ref descriptor 1))
X
X(define (vm-set-procedure-code! descriptor code)
X  (assert (vm-procedure? descriptor) 'vm-set-procedure-code!)
X  (stored-object-set! descriptor 0 code))
X
X(define (vm-set-procedure-environment! descriptor environment)
X  (assert (vm-procedure? descriptor) 'vm-set-procedure-environment!)
X  (stored-object-set! descriptor 1 environment))
X
X(define (vm-make-procedure)
X  (allocate-stored-object procedure-tag 2))
X
X;;; THE STACK
X
X; The stack grows toward low memory, and the stack pointer always points at the
X; location from which a POP will fetch.
X
X(define *sp* (make-register 0))
X
X(define (reset-stack-pointer)
X  (store! *sp* (area-end *stack-area*)))
X
X(define stack-allocation-counter (make-counter "Stack allocated"))
X
X(define (push descriptor)
X  (if (<= (fetch *sp*) (area-begin *stack-area*))
X      (error "stack overflow" descriptor))
X  (store! *sp* (- (fetch *sp*) 1))
X  (add-to-counter stack-allocation-counter 1)
X  (vector-set! *memory* (fetch *sp*) descriptor))
X
X(define (pop)
X  (let ((descriptor (vector-ref *memory* (fetch *sp*))))
X    (store! *sp* (+ (fetch *sp*) 1))
X    descriptor))
X
X(define (make-stack-vector len)
X  (push (vm-make-small-integer len))
X  (make-stored-object-descriptor vector-tag (fetch *sp*)))
X
X
X
X;;; VIRTUAL MACHINE: INTERPRETER
X
X(define *val* (make-descriptor-register))
X(define *env* (make-descriptor-register))
X(define *cont* (make-descriptor-register))
X(define *code* (make-descriptor-register))
X(define *pc* (make-register -1))
X(define *nargs* (make-register -1))
X
X(define (push-state pc size)
X  (push (fetch *env*))
X  (push (vm-make-small-integer pc))
X  (push (fetch *code*))
X  (push (fetch *cont*))
X  (store! *cont*
X          (make-stack-vector (+ size 4))))
X
X(define (pop-state cont)
X  (store! *sp* (+ (address-of-length-cell cont) 1))
X  (store! *cont* (pop))
X  (store! *code* (pop))
X  (store! *pc* (vm-small-integer-value (pop)))
X  (store! *env* (pop)))
X
X(define (next-instruction-word)
X  (let ((word (vm-vector-ref (fetch *code*) (fetch *pc*))))
X    (store! *pc* (+ (fetch *pc*) 1))
X    word))
X
X(define (next-instruction-integer)
X  (vm-small-integer-value (next-instruction-word)))
X
X(define *trace?* #f)
X
X(define (interpret)
X  (let interpret-loop ()
X    (let ((op (next-instruction-integer)))
X      (if *trace?* (begin (newline) (write (opcode-name op))))
X      ((vector-ref instruction-dispatch-vector op))
X      (if (>= (fetch *pc*) 0)
X          (interpret-loop)))))
X
X(define instruction-dispatch-vector
X  (make-vector number-of-opcodes (lambda () (error "unimplemented opcode"))))
X
X(define (define-opcode name procedure)
X  (vector-set! instruction-dispatch-vector (opcode name) procedure))
X
X(define-opcode 'check-nargs
X  (lambda ()
X    (let ((desired-nargs (next-instruction-integer)))
X      (if (not (= (fetch *nargs*) desired-nargs))
X          (error "wrong number of arguments" desired-nargs (fetch *nargs*))))))
X
X(define-opcode 'make-environment
X  (lambda ()
X    (let ((nargs (next-instruction-integer)))
X      (let ((v (vm-make-vector (+ nargs 1))))
X        (vm-vector-set! v 0 (fetch *env*))
X        (do ((i 1 (+ i 1)))
X            ((> i nargs) (store! *env* v))
X          (vm-vector-set! v i (pop)))))))
X
X(define-opcode 'load-constant
X  (lambda ()
X    (store! *val* (next-instruction-word))))
X
X(define (nth-outer-rib n)
X  (let loop ((n n) (rib (fetch *env*)))
X    (if (<= n 0) rib (loop (- n 1) (vm-vector-ref rib 0)))))
X
X(define-opcode 'load-variable
X  (lambda ()
X    (let ((rib (nth-outer-rib (next-instruction-integer))))
X      (let ((val (vm-vector-ref rib (next-instruction-integer))))
X        (if (same-descriptor? val vm-undefined)
X            (error "undefined variable"))
X        (store! *val* val)))))
X
X(define-opcode 'set-variable
X  (lambda ()
X    (let ((rib (nth-outer-rib (next-instruction-integer))))
X      (vm-vector-set! rib (next-instruction-integer) (fetch *val*)))))
X
X(define-opcode 'make-procedure
X  (lambda ()
X    (let ((proc (vm-make-procedure)))
X      (store! *val* proc)
X      (vm-set-procedure-code! proc (next-instruction-word))
X      (vm-set-procedure-environment! proc (fetch *env*)))))
X
X(define-opcode 'push
X  (lambda ()
X    (push (fetch *val*))))
X
X(define-opcode 'call
X  (lambda ()
X    (let ((nargs (next-instruction-integer)))
X;      (assert (= (+ (fetch *sp*) nargs)
X;                 (address-of-length-cell (fetch *cont*)))
X;              'call)
X      (start-call (fetch *val*) nargs))))
X
X(define (start-call proc nargs) (store! *nargs* nargs) (store! *code*
X(vm-procedure-code proc)) (store! *env* (vm-procedure-environment
Xproc)) (store! *pc* 0))
X
X(define-opcode 'make-continuation
X  (lambda ()
X    (let ((return-pc (next-instruction-integer)))
X      (push-state return-pc
X                  (next-instruction-integer)))))
X
X(define-opcode 'return
X  (lambda ()
X    (pop-state (fetch *cont*))))
X
X(define-opcode 'leave-environment
X  (lambda ()
X    (store! *env* (vm-vector-ref (fetch *env*) 0))))
X
X(define-opcode 'jump-if-false
X  (lambda ()
X    (let ((offset (next-instruction-integer)))
X      (if (not (vm-boolean-value (fetch *val*)))
X          (store! *pc* offset)))))
X
X(define-opcode 'jump
X  (lambda ()
X    (store! *pc* (next-instruction-integer))))
X
X
X
X; Handy combinators for making primitives
X
X(define (no-arg-primitive proc)
X  (lambda ()
X    (store! *val* (proc))))
X
X(define (one-arg-primitive proc)
X  (lambda ()
X    (store! *val* (proc (fetch *val*)))))
X
X(define (two-arg-primitive proc)
X  (lambda ()
X    (store! *val* (proc (fetch *val*) (pop)))))
X
X(define (three-arg-primitive proc)
X  (lambda ()
X    (let ((second-arg (pop)))
X      (store! *val* (proc (fetch *val*) second-arg (pop))))))
X
X(define (one-arg-predicate-primitive proc)
X  (one-arg-primitive (lambda (d) (vm-make-boolean (proc d)))))
X
X(define (two-arg-predicate-primitive proc)
X  (two-arg-primitive (lambda (d1 d2) (vm-make-boolean (proc d1 d2)))))
X
X(define (no-arg-effect-primitive proc)
X  (no-arg-primitive (lambda () (proc) vm-unspecified)))
X
X(define (one-arg-effect-primitive proc)
X  (one-arg-primitive (lambda (d) (proc d) vm-unspecified)))
X
X(define (two-arg-effect-primitive proc)
X  (two-arg-primitive (lambda (d1 d2) (proc d1 d2) vm-unspecified)))
X
X(define (arithmetic-primitive proc)
X  (two-arg-primitive
X   (lambda (d1 d2)
X     (vm-make-small-integer (proc (vm-small-integer-value d1)
X                                  (vm-small-integer-value d2))))))
X
X(define (arithmetic-comparison-primitive proc)
X  (two-arg-predicate-primitive (lambda (d1 d2)
X                                 (proc (vm-small-integer-value d1)
X                                       (vm-small-integer-value d2)))))
X
X; Primitives
X
X(define-opcode 'eq?
X  (two-arg-predicate-primitive same-descriptor?))
X
X(define-opcode 'integer?
X  (one-arg-predicate-primitive vm-small-integer?))
X
X(define-opcode '+         (arithmetic-primitive +))
X(define-opcode '-          (arithmetic-primitive -))
X(define-opcode '*          (arithmetic-primitive *))
X(define-opcode 'quotient  (arithmetic-primitive quotient))
X(define-opcode 'remainder (arithmetic-primitive remainder))
X
X(define-opcode '= (arithmetic-comparison-primitive =))
X(define-opcode '< (arithmetic-comparison-primitive <))
X
X(define-opcode 'pair? (one-arg-predicate-primitive vm-pair?))
X
X(define-opcode 'cons
X  (lambda ()
X    (let ((pair (vm-make-pair)))
X      (vm-set-car! pair (fetch *val*))
X      (vm-set-cdr! pair (pop))
X      (store! *val* pair))))
X
X(define-opcode 'car (one-arg-primitive vm-car))
X(define-opcode 'cdr (one-arg-primitive vm-cdr))
X(define-opcode 'set-car! (two-arg-effect-primitive vm-set-car!))
X(define-opcode 'set-cdr! (two-arg-effect-primitive vm-set-cdr!))
X
X(define-opcode 'ref? (one-arg-predicate-primitive vm-ref?))
X(define-opcode 'ref
X  (lambda ()
X    (let ((the-ref (vm-make-ref)))
X      (vm-put! the-ref (fetch *val*))
X      (store! *val* the-ref))))
X(define-opcode 'get (one-arg-primitive vm-get))
X(define-opcode 'put! (two-arg-effect-primitive vm-put!))
X
X(define-opcode 'vector? (one-arg-predicate-primitive vm-vector?))
X
X(define-opcode 'make-vector
X  (lambda ()
X    (let ((vec (vm-make-vector (vm-small-integer-value (fetch *val*)))))
X      (vm-vector-fill! vec (pop))
X      (store! *val* vec))))
X
X(define-opcode 'vector-length (one-arg-primitive vm-vector-length))
X
X(define-opcode 'vector-ref
X  (two-arg-primitive (lambda (d1 d2)
X                       (vm-vector-ref d1 (vm-small-integer-value d2)))))
X
X(define-opcode 'vector-set!
X  (three-arg-primitive (lambda (d1 d2 d3)
X                         (vm-vector-set! d1 (vm-small-integer-value d2) d3)
X                         vm-unspecified)))
X
X(define-opcode 'read
X  (no-arg-primitive (lambda () (copy-to-heap (read)))))
X
X(define-opcode 'write
X  (one-arg-effect-primitive (lambda (d)
X                              (write (extract-from-heap d)))))
X
X(define-opcode 'display 
X  (one-arg-effect-primitive (lambda (d)
X                              (display (extract-from-heap d)))))
X
X(define-opcode 'newline
X  (no-arg-effect-primitive newline))
X
X
X
X;;; Maybe should redo opcodes here just in case.
X(define (initialize)
X  (clear-descriptor-registers)
X  (initialize-memory (+ (* 2 *heap-size*) *stack-size*))
X  (initialize-areas)
X  (reset-stack-pointer)
X  (reset-symbol-table))
X
X; Entry to interpreter
X
X(define *global-env* (make-descriptor-register))
X(define *global-env-size* 200)
X
X(define (validate-global-environment)
X  (cond ((not (vm-vector? (fetch *global-env*)))
X         (set! *global-variables* '())
X         (store! *global-env*
X                 (vm-make-vector *global-env-size*))
X         (vm-vector-fill! (fetch *global-env*) vm-undefined))))
X
X(define (vm-define var descriptor)
X  (vm-vector-set! (fetch *global-env*)
X                  (global-variable-index var)
X                  descriptor))
X
X(define (run exp)                        ;don't forget to (initialize)
X  (run! exp)
X  (extract-from-heap (fetch *val*)))
X
X(define *show-counters?* #t)
X
X(define (run! exp)
X  (let ((code (compile `(lambda () ,exp))))
X    (reset-stack-pointer)                ;in case it overflowed
X    (validate-global-environment)
X    (push (copy-to-heap code))
X    (let ((proc (vm-make-procedure)))
X      (vm-set-procedure-code! proc (pop))
X      (vm-set-procedure-environment! proc (fetch *global-env*))
X      (reset-counters)
X      (start-interpreter proc)
X      (if *show-counters?* (show-counters)))))
X
X(define (start-interpreter thunk)
X  (push-state -1 0)
X  (start-call thunk 0)
X  (interpret))
X
X
X
X; Read-evaluate-print loop.
X
X(define (repl)                                ;don't forget to (initialize)
X  (let loop ()
X    (newline)
X    (display "PLScheme==> ")
X    (let ((form (read)))
X      (cond ((definition? form)
X             (newline)
X             (let ((var (definition-lhs form))
X                   (exp (definition-rhs form)))
X               (run! exp)
X               (vm-define var (fetch *val*))
X               (write var)
X               (display " defined.")))
X            (else
X             (newline)
X             (run! form)
X             (write (extract-from-heap (fetch *val*))))))
X    (loop)))
X
X; Test routines.
X; You must do (initialize) before calling these.
X
X(define (test-fact n)
X  (run `(letrec ((fact (lambda (n) (if (< n 2) n (* n (fact (- n 1)))))))
X          (fact ,n))))
X
X(define (test-ifact n)
X  (run `(let ((ifact (lambda (n)
X                       (let loop ((i n) (a 1))
X                         (if (< i 2) a (loop (- i 1) (* i a)))))))
X          (ifact ,n))))
X
X(define (cons-a-lot)  ;This exercises the garbage collector.
X  (run '(let loop ((n 2000))
X          (if (< n 0)
X              'done
X              (begin (cons 'this 'that) (loop (- n 1)))))))
X
X
X
X;;; VIRTUAL MACHINE: GARBAGE COLLECTOR
X
X(define *gc-noisily?* #t)
X
X(define (collect)
X  (if *gc-noisily?* (begin (newline) (write 'gc)))
X  (let ((before (- (area-end *current-area*)
X                   (next-available *current-area*))))
X    (flip)
X    (scan-registers)
X    (scan-stack)
X    (scan-transitively)
X    (discard)
X    (let ((after (- (area-end *current-area*)
X                    (next-available *current-area*))))
X      (if *gc-noisily?*
X          (begin (display " reclaimed ")
X                 (write (- after before))
X                 (newline))))))
X
X(define (flip)
X  (let ((full-area *current-area*))
X    (set! *current-area* *other-area*)
X    (set! *other-area* full-area)))
X
X(define (discard)
X  (set-next-available! *other-area* (area-begin *other-area*)))
X
X(define (scan-registers)
X  (for-each (lambda (register)
X              (store! register (maybe-copy (fetch register))))
X            *descriptor-registers*))
X
X(define (scan-stack)
X  (let ((end (area-end *stack-area*)))
X    (let scan-loop ((scan-pointer (fetch *sp*)))
X      (cond ((< scan-pointer end)
X             (vector-set! *memory*
X                          scan-pointer
X                          (maybe-copy (vector-ref *memory*
X                                                  scan-pointer)))
X             (scan-loop (+ scan-pointer 1)))))))
X
X(define (scan-transitively)
X  (let scan-loop ((scan-pointer (area-begin *current-area*)))
X    (cond ((< scan-pointer (next-available *current-area*))
X           (vector-set! *memory*
X                        scan-pointer
X                        (maybe-copy (vector-ref *memory* scan-pointer)))
X           (scan-loop (+ scan-pointer 1))))))
X
X
X
X(define (maybe-copy descriptor)
X  (cond ((not (stored-object? descriptor)) descriptor)
X        ((forwarded? descriptor) (forwarded-to descriptor))
X        ((in-area? descriptor *other-area*)
X         (let ((len (stored-object-length descriptor)))
X           (let ((new (allocate-stored-object (descriptor-tag descriptor)
X                                              len)))
X             (do ((i 0 (+ i 1)))
X                 ((>= i len))
X               (stored-object-set! new
X                                   i
X                                   (stored-object-ref descriptor i)))
X             (set-forwarded-to! descriptor new)
X             new)))
X        (else descriptor)))
X
X
X
X;;; VIRTUAL MACHINE: BOOTSTRAPPING AND DEBUGGING
X
X(define (copy-to-heap object)
X  (cond ((integer? object) (vm-make-small-integer object))
X        ((null? object) vm-empty-list)
X        ((equal? object true) vm-true)
X        ((equal? object false) vm-false)
X        ((equal? object unspecified) vm-unspecified)
X        ((pair? object)
X         (push (copy-to-heap (car object)))
X         (push (copy-to-heap (cdr object)))
X         (let ((pair (vm-make-pair)))
X           (vm-set-cdr! pair (pop))
X           (vm-set-car! pair (pop))
X           pair))
X        ((vector? object)
X         (let ((len (vector-length object)))
X           (let ((v (vm-make-vector len)))
X             (vm-vector-fill! v vm-unspecified)
X             (push v))
X           (do ((i 0 (+ i 1)))
X               ((>= i len) (pop))
X             (let ((component (copy-to-heap (vector-ref object i))))
X               (let ((v (pop)))
X                 (vm-vector-set! v i component)
X                 (push v))))))
X        ((symbol? object)
X         (find-or-make-symbol object))
X        (else
X         (error "don't know how to copy this into heap"
X                object))))
X
X
X
X(define (extract-from-heap descriptor)
X  (let loop ((descriptor descriptor)
X              (trail '()))                ;don't blow up on circular objects
X    (if (member descriptor trail)
X        '<cycle>
X        (let ((trail (cons descriptor trail)))
X          (cond ((vm-small-integer? descriptor)
X                 (vm-small-integer-value descriptor))
X                ((vm-empty-list? descriptor) '())
X                ((same-descriptor? descriptor vm-false) false)
X                ((same-descriptor? descriptor vm-true) true)
X                ((same-descriptor? descriptor vm-unspecified) unspecified)
X                ((same-descriptor? descriptor (fetch *global-env*))
X                 '<global-env>)
X                ((vm-pair? descriptor)
X                 (cons (loop (vm-car descriptor) trail)
X                       (loop (vm-cdr descriptor) trail)))
X                ((vm-vector? descriptor)
X                 (let ((len (vm-vector-length descriptor)))
X                   (let ((v (make-vector len)))
X                     (do ((i 0 (+ i 1)))
X                         ((>= i len) v)
X                       (vector-set! v i
X                                    (loop (vm-vector-ref descriptor i)
X                                           trail))))))
X                ((vm-symbol? descriptor)
X                 (extract-symbol-from-heap descriptor))
X                ((vm-ref? descriptor)
X                 ;; extract?
X                 (vector '<ref> (loop (vm-get descriptor) trail)))
X                ((vm-procedure? descriptor)
X                 (vector '<procedure>
X                         (loop (vm-procedure-code descriptor) trail)
X                         (loop (vm-procedure-environment descriptor) trail)))
X                (else
X                 (vector 'unknown-type descriptor)))))))
X
END_OF_FILE
if test 28360 -ne `wc -c <'plscheme/vm.scm'`; then
    echo shar: \"'plscheme/vm.scm'\" unpacked with wrong size!
fi
# end of 'plscheme/vm.scm'
fi
echo shar: End of archive 2 \(of 2\).
cp /dev/null ark2isdone
MISSING=""
for I in 1 2 ; do
    if test ! -f ark${I}isdone ; then
	MISSING="${MISSING} ${I}"
    fi
done
if test "${MISSING}" = "" ; then
    echo You have unpacked both archives.
    rm -f ark[1-9]isdone
else
    echo You still need to unpack the following archives:
    echo "        " ${MISSING}
fi
##  End of shell archive.
exit 0
